"""A collection of parameters jointly defining an operating or optimization space"""

from .base import (
    IParamSpace,
    Parameter
)
from .continuous import (
    Param_Continuous,
    Param_Observational
)
from .discrete import (
    Param_Categorical,
    Param_Ordinal,
    Param_Discrete,
    Param_Discrete_Numeric,
    Task,
    CAT_SEP
)

from .config import param_class_dict
from obsidian.constraints import (
    Input_Constraint,
    Linear_Constraint,
    Nonlinear_Constraint,
    const_class_dict
)

from obsidian.exceptions import UnsupportedError
from obsidian.utils import tensordict_to_dict

import torch
import numpy as np
import pandas as pd


class ParamSpace(IParamSpace):
    """
    Class designed to define the parameter space in which an optimization
    can be conducted.

    Attributes:
        params (tuple[Parameter]): A tuple of Parameter types defining the parameter space.
        X_names (tuple[str]): A tuple of the names of the parameters.
        X_cont (list[str]): A list of the names of the continuous parameters.
        X_obs (list[str]): A list of the names of the observational parameters.
        X_discrete (dict[str, list[str]]): A dictionary mapping the names of the discrete parameters to their categories.
        X_task (dict[str, list[str]]): A dictionary mapping the names of the task parameters to their categories.
        X_min (torch.Tensor): A tensor containing the minimum values of the parameters.
        X_max (torch.Tensor): A tensor containing the maximum values of the parameters.
        X_range (torch.Tensor): A tensor containing the range of the parameters.
        X_static (list[str]): A list of the names of the static parameters.
        X_vary (list[str]): A list of the names of the varying parameters.
        n_dim (int): The number of dimensions in the parameter space.
        n_tdim (int): The total number of dimensions in the parameter space after transformation.
        t_map (dict): A dictionary mapping parameter indices to transformed indices.
        tinv_map (dict): A dictionary mapping transformed indices to parameter indices.
        X_discrete_idx (list[int]): A list of the indices of the discrete parameters.
        X_t_discrete_idx (list[int]): A list of the indices of the transformed discrete parameters.
        X_t_cat_idx (list[int]): A list of the indices of the transformed categorical parameters.
        X_t_task_idx (int): The index of the transformed task parameter.
        search_space (pd.DataFrame): The allowable search space for future optimization.
    
    Raises:
        ValueError: If the X_names are not unique.
        UnsupportedError: If there is more than one task parameter.

    """
    def __init__(self,
                 params: list[Parameter]):
        
        # Convert to immutable dtype to presever order
        self.params = tuple(params)
        self.X_names = tuple([x.name for x in self])
        if len(set(self.X_names)) != len(self.X_names):
            raise ValueError("X_names must be unique.")
        if any([CAT_SEP in x for x in self.X_names]):
            raise ValueError(f"X_names cannot contain '{CAT_SEP}'.")

        # Save as attributes key types that affect usage behavior
        self.X_cont = [x for x in self if isinstance(x, Param_Continuous)
                       and not isinstance(x, Param_Observational)]
        self.X_obs = [x for x in self if isinstance(x, Param_Observational)]
        self.X_discrete = [x for x in self if isinstance(x, Param_Discrete)]
        self.X_task = [x for x in self if isinstance(x, Task)]
        if len(self.X_task) > 1:
            raise UnsupportedError("Only one task parameter is allowed.")
        
        # Locally calculate other types for counting

        X_cat = [x for x in self if isinstance(x, Param_Categorical)]
        X_ord = [x for x in self if isinstance(x, Param_Ordinal)]
        X_disc_num = [x for x in self if isinstance(x, Param_Discrete_Numeric)]
        
        # Extract ranges for scaling
        self.X_min = torch.Tensor([x.min for x in self])
        self.X_max = torch.Tensor([x.max for x in self])
        self.X_range = self.X_max - self.X_min

        # Extract where variables are changing or are fixed/observational
        self.X_static = [x.name for x in self if x.min == x.max or isinstance(x, Param_Observational)]
        self.X_vary = [name for name in self.X_names if name not in self.X_static]

        # Extract number of dimensions (raw input and encoded)
        self.n_dim = len(self.X_names)
        # Total number of dimensions after transform
        # Numerical + Observational + Unique Categories + Ordinal + Discrete_Numeric
        self.n_tdim = int(len(self.X_cont)
                          + len(self.X_obs)
                          + np.sum([len(x.categories) for x in X_cat])
                          + len(X_ord)
                          + len(X_disc_num))
        
        # Assign the inv/transform maps
        self.t_map = self.map_transform()
        self.tinv_map = self.map_inv_transform()

        # Save the locations of discrete variables
        self.X_discrete_idx = [i for i in self.t_map.keys() if self[i] in self.X_discrete]
        self.X_t_discrete_idx = [i_t for i_t, i in self.tinv_map.items() if self[i] in self.X_discrete]
        self.X_t_cat_idx = [i_t for i_t, i in self.tinv_map.items() if self[i] in X_cat]
        if self.X_task:
            self.X_t_task_idx = next(i_t for i_t, i in self.tinv_map.items() if self[i] in self.X_task)
        else:
            self.X_t_task_idx = None
        
        # Set up storage for constraints
        self.linear_constraints = []
        self.nonlinear_constraints = []

        return

    def map_transform(self) -> dict:
        """
        Maps the parameter indices to transformed indices based on the parameter types.

        Returns:
            dict: A dictionary mapping parameter indices to transformed indices.
        """
        t_map = {}
        count = 0
        for i, param_i in enumerate(self.params):
            if isinstance(param_i, Param_Categorical):
                t_map[i] = [count + j for j in range(len(param_i.categories))]
                count += len(param_i.categories) - 1
            else:
                t_map[i] = count
            count += 1

        return t_map
    
    def map_inv_transform(self) -> dict:
        """
        Maps the inverse of the transformed dictionary.

        Returns:
            dict: A dictionary where the keys are the original values and the values are the original keys.
        """
        t_map = self.map_transform()
        tinv_map = {}
        for k, v in t_map.items():
            if isinstance(v, list):
                for i in v:
                    tinv_map[i] = k
            else:
                tinv_map[v] = k

        return tinv_map
    
    # Establish a handler method to transform X based on type
    def _transform(self,
                   X: pd.DataFrame,
                   type: str) -> pd.DataFrame:
        """
        Transforms the input DataFrame `X` based on the specified `type`.

        Args:
            X (pd.DataFrame): The input DataFrame to be transformed.
            type (str): The type of transformation to be applied.

        Returns:
            pd.DataFrame: The transformed DataFrame.

        Raises:
            KeyError: If an unknown column is passed in `X`.
        """
        cols = X.columns
        # Check for column match, but consider that categorical OH encoding will lead to Parameter N_i type names
        if not all(col.split(CAT_SEP)[0] in self.X_names for col in cols):
            raise KeyError('Unknown column passed in X')
        X_t = pd.DataFrame()
        for param_i in self:
            methods = {'unit_map': param_i.unit_map,
                       'unit_demap': param_i.unit_demap,
                       'encode': param_i.encode,
                       'decode': param_i.decode}
            # Include all matches, including OH columns (e.g. Parameter N_i)
            cols = [col for col in X.columns if param_i.name in col]
            # Skip if one of the possible parameters is excluded from the transform operation
            if cols != []:
                X_t_i = pd.DataFrame(methods[type](X[cols].values))
                X_t_i.columns = [param_i.name] if not (
                    isinstance(param_i, Param_Categorical) and type == 'encode') else X_t_i.columns
                X_t = pd.concat([X_t, X_t_i], axis=1)

        if type in ['encode', 'unit_map']:
            X_t = X_t.apply(pd.to_numeric)  # Do not send object dtypes for these transforms

        return X_t

    # Define transformation methods using a handler method and lambdas; more readable on outer layer
    def unit_map(self, X):
        """Map from measured to 0,1 space"""
        return self._transform(X, type='unit_map')
    
    def unit_demap(self, X):
        """Map from 0,1 space to measured space"""
        return self._transform(X, type='unit_demap')
    
    def encode(self, X):
        """Encode parameter to a format that can be used for training"""
        return self._transform(X, type='encode')

    def decode(self, X):
        """Decode parameter from transformed space"""
        return self._transform(X, type='decode')

    @property
    def search_space(self) -> pd.DataFrame:
        """
        Returns the search space for the parameter space.

        Returns:
            pd.DataFrame: A dataframe containing the search space for the parameter space.
        """
        
        # Establish the boundaries in real space
        X_search_t = pd.DataFrame()
        for param in self:
            # For continuous, encode the continuous bounds
            if isinstance(param, Param_Continuous):
                cont_bounds = pd.DataFrame(param.encode([param.search_min, param.search_max]), columns=[param.name])
                X_search_t = pd.concat([X_search_t, cont_bounds], axis=1)
                
            # For discrete, encode the available categories, then log the min-max of encoded columns
            elif isinstance(param, Param_Discrete) and (not isinstance(param, (Param_Ordinal, Param_Discrete_Numeric))):
                # Discrete parameter bounds aren't actually handled here; they are handled in optimizer._fixed_features())
                cat_e = param.encode(param.search_categories)
                disc_bounds = pd.DataFrame(np.vstack([[0]*cat_e.shape[-1], cat_e.max().values]),
                                           columns=cat_e.columns)
                X_search_t = pd.concat([X_search_t, disc_bounds], axis=1)
            elif isinstance(param, (Param_Ordinal, Param_Discrete_Numeric)):
                cat_e = param.encode(param.search_categories)
                cont_bounds = pd.DataFrame([min(cat_e), max(cat_e)], columns=[param.name])
                X_search_t = pd.concat([X_search_t, cont_bounds], axis=1)
        
        return X_search_t

    def open_search(self) -> None:
        """Set the search space to the parameter space"""
        for param in self:
            param.open_search()
    
    def mean(self) -> pd.DataFrame:
        """
        Calculates the mean values for each parameter in the parameter space.

        Returns:
            pd.DataFrame: A DataFrame containing the mean values for each parameter.
        """
        row = {}
        for param_i in self:
            if isinstance(param_i, Param_Continuous):  # Mean of continuous
                row[param_i.name] = param_i.unit_demap([0.5])[0]
            elif isinstance(param_i, Param_Discrete):  # First of discrete
                row[param_i.name] = param_i.categories[0]

        df_mean = pd.DataFrame([row])

        return df_mean

    def constrain_inputs(self,
                         constraint: Input_Constraint) -> None:
        """
        Constrains the input space based on the specified equality, inequality, or nonlinear constraint.

        Args:
            constraint (Input_Constraint): The constraint to be applied to the input space.
        """
        if isinstance(constraint, Linear_Constraint):
            self.linear_constraints.append(constraint)
        elif isinstance(constraint, Nonlinear_Constraint):
            self.nonlinear_constraints.append(constraint)
        
        return

    def clear_constraints(self) -> None:
        """Clears all constraints from the input space."""
        self.linear_constraints = []
        self.nonlinear_constraints = []
        return

    def save_state(self) -> dict:
        """
        Saves the state of the ParamSpace object.

        Returns:
            dict: A dictionary containing the state of the ParamSpace object.
        """
        obj_dict = {}
        for param in self:
            obj_dict[param.name] = {'class': param.__class__.__name__,
                                    'state': param.save_state()}
            
        if getattr(self, 'linear_constraints', []):
            obj_dict['linear_constraints'] = [{'class': const.__class__.__name__,
                                               'state': tensordict_to_dict(const.state_dict())}
                                              for const in self.linear_constraints]
        if getattr(self, 'nonlinear_constraints', []):
            obj_dict['nonlinear_constraints'] = [{'class': const.__class__.__name__,
                                                  'state': tensordict_to_dict(const.state_dict())}
                                                 for const in self.nonlinear_constraints]

        return obj_dict
    
    @classmethod
    def load_state(cls,
                   obj_dict: dict):
        """
        Loads the state of the ParamSpace object from a dictionary.

        Args:
            obj_dict (dict): A dictionary containing the state of the ParamSpace object.

        Returns:
            ParamSpace: A new ParamSpace object with the loaded state.

        """

        params = []
        for param, param_dict in obj_dict.items():
            if 'constraint' not in param:
                param = param_class_dict[param_dict['class']].load_state(param_dict['state'])
                params.append(param)

        new_X_space = cls(params=params)

        if 'linear_constraints' in obj_dict:
            for const_dict in obj_dict['linear_constraints']:
                const = const_class_dict[const_dict['class']](new_X_space, **const_dict['state'])
                new_X_space.constrain_inputs(const)
        if 'nonlinear_constraints' in obj_dict:
            for const_dict in obj_dict['nonlinear_constraints']:
                const = const_class_dict[const_dict['class']](new_X_space, **const_dict['state'])
                new_X_space.constrain_inputs(const)

        return new_X_space



"""PyTests for obsidian.plotting"""

from obsidian import Campaign
from obsidian.plotting import (
    parity_plot,
    factor_plot,
    surface_plot,
    visualize_inputs,
    optim_progress,
    MDS_plot
)

from obsidian.objectives import Scalar_WeightedSum

import pytest
from obsidian.tests.utils import DEFAULT_MOO_PATH
import json

# Avoid using TkAgg which causes Tcl issues during testing
import matplotlib
matplotlib.use('inline')

with open(DEFAULT_MOO_PATH) as json_file:
    obj_dict = json.load(json_file)

campaign = Campaign.load_state(obj_dict)
optimizer = campaign.optimizer


@pytest.mark.fast
def test_parity_plot():
    fig = parity_plot(optimizer)


plots_and_objects = [pytest.param(parity_plot, optimizer, marks=pytest.mark.fast),
                     pytest.param(visualize_inputs, campaign, marks=pytest.mark.fast),
                     pytest.param(factor_plot, optimizer, marks=pytest.mark.fast),
                     (surface_plot, optimizer)]


@pytest.mark.parametrize('plot_type, acting_object', plots_and_objects)
def test_plots_fast(plot_type, acting_object):
    fig = plot_type(acting_object)


feature_ids = [i for i in range(len(optimizer.X_space))]


@pytest.mark.parametrize('feature_id', feature_ids)
def test_factor_plot(feature_id):
    fig = factor_plot(optimizer, feature_id=feature_id)


def test_factor_plot_options():
    X_ref = optimizer.X_train.iloc[-1, :].to_frame().T
    fig = factor_plot(optimizer, f_transform=True, X_ref=X_ref)


@pytest.mark.slow
@pytest.mark.parametrize('feature_id', feature_ids)
def test_surface_plot(feature_id):
    fig = surface_plot(optimizer, feature_ids=[0, feature_id])


@pytest.mark.slow
def test_surface_plot_options():
    fig = surface_plot(optimizer, plot_bands=False, plot_data=True)
    fig = surface_plot(optimizer, f_transform=True)


@pytest.mark.fast
def test_optim_progress_plot():
    campaign.clear_objective()
    X_suggest, eval_suggest = campaign.suggest(optim_samples=2, optim_restarts=1)
    fig = optim_progress(campaign, X_suggest=X_suggest)
    obj = Scalar_WeightedSum(weights=[1, 1])
    campaign.set_objective(obj)
    fig = optim_progress(campaign, X_suggest=X_suggest)
    
    
@pytest.mark.fast
def test_MDS_plot():
    # Test with default data
    fig = MDS_plot(campaign)
    
    # Test with no iteration data
    data = campaign.data
    data = data.drop(columns='Iteration')
    campaign.clear_data()
    campaign.add_data(data)
    fig = MDS_plot(campaign)


if __name__ == '__main__':
    pytest.main([__file__, '-m', 'not slow'])



"""Utility functions for PyTest unit testing"""

from obsidian.tests.param_configs import X_sp_default, X_sp_cont_ndims

from obsidian import Campaign, ParamSpace, Target
from obsidian.experiment import Simulator
from obsidian.experiment.benchmark import shifted_parab, two_leaves
from obsidian.objectives import Identity_Objective

from numpy.typing import ArrayLike
from typing import Callable

import pandas as pd
import numpy as np
import json

# Default campaigns for testing without having to re-run optimization
DEFAULT_MOO_PATH = 'obsidian/tests/default_campaign_MOO.json'
DEFAULT_SOO_PATH = 'obsidian/tests/default_campaign_SOO.json'


def equal_state_dicts(e1: dict | str | float | int,
                      e2: dict | str | float | int) -> bool:
    """
    Recursively compare two dictionaries, and allow for floating-point error
    """
    # If the values are equal, skip
    if not e1 == e2:
        # First, make sure we are comparing the same type
        assert type(e1) == type(e2), f'Type mismatch at {e1} != {e2}'
        
        # If they are dictionaries, compare elements recursively
        if isinstance(e1, dict):
            assert e1.keys() == e2.keys(), f'Keys mismatch at {e1.keys()} != {e2.keys()}'
            for k, v in e1.items():
                equal_state_dicts(v, e2[k])
                
        # If they are lists, compare elements recursively
        elif isinstance(e1, list):
            assert len(e1) == len(e2), f'Length mismatch at {len(e1)} != {len(e2)}'
            for e1_i, e2_i in zip(e1, e2):
                equal_state_dicts(e1_i, e2_i)
                
        # Otherwise, if it is numerical, check for floating-point error
        elif isinstance(e1, (float, int)):
            if not (np.isnan(e1) and np.isnan(e2)):
                assert (e1-e2)/e1 < 1e-6, f'{e1} != {e2}'
        else:
            raise ValueError(f'{e1} != {e2}')
        
    return True


def approx_equal(x1: ArrayLike | float | int,
                 x2: ArrayLike | float | int,
                 tol: float = 1e-6):
    """
    Check if two numbers or arrays are approximately equal within a given tolerance.

    Args:
        x1 (ArrayLike or numeric): The first number.
        x2 (ArrayLike or numeric): The second number.
        tol (float, optional): The tolerance value. Defaults to 1e-6.

    Returns:
        bool: True if the numbers are approximately equal within the tolerance,
            False otherwise.
    """
    diff = abs((x1-x2)/x1)
    return diff.max() < tol


def save_default_campaign(X_space: ParamSpace,
                          response_func: Callable,
                          path: str,
                          m_initial: int = 6,
                          n_response: int = 1):  # pragma: no cover
    
    y_name = 'Response'
    y_names = [y_name+' '+str(i+1) for i in range(n_response)]
    target = [Target(name=n, f_transform='Standard', aim='max') for n in y_names]
    campaign = Campaign(X_space, target)
    simulator = Simulator(X_space, response_func, y_names)
    
    X0 = campaign.designer.initialize(m_initial=m_initial, method='LHS')
    y0 = simulator.simulate(X0)
    Z0 = pd.concat([X0, y0], axis=1)
    campaign.add_data(Z0)

    for i in range(3):
        print(f'Simulating iteration {i+1}')
        campaign.fit()
        
        X_suggest, eval_suggest = campaign.optimizer.suggest(
            m_batch=2, objective=Identity_Objective(mo=n_response > 1))
        y_i = simulator.simulate(X_suggest)
        Z_i = pd.concat([X_suggest, y_i, eval_suggest], axis=1)
        campaign.add_data(Z_i)
    
    obj_dict = campaign.save_state()
    with open(path, 'w') as outfile:
        json.dump(obj_dict, outfile)


if __name__ == '__main__':
    
    save_default_campaign(X_sp_default, shifted_parab, DEFAULT_SOO_PATH, n_response=1)
    save_default_campaign(X_sp_cont_ndims[2], two_leaves, DEFAULT_MOO_PATH, n_response=2)



"""Surrogate models built using BoTorch API and torch_model objects"""

from .base import SurrogateModel
from .config import model_class_dict
from .utils import fit_pytorch

from obsidian.utils import tensordict_to_dict, dict_to_tensordict
from obsidian.exceptions import SurrogateFitError
from obsidian.config import TORCH_DTYPE

from botorch.fit import fit_gpytorch_mll
from botorch.optim.fit import fit_gpytorch_mll_torch, fit_gpytorch_mll_scipy
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.ensemble import EnsembleModel
from gpytorch.mlls import ExactMarginalLogLikelihood

import torch
import torch.nn as nn
import numpy as np
import pandas as pd
import warnings


class SurrogateBoTorch(SurrogateModel):
    """
    BoTorch GP model, subclass of the obsidian SurrogateModel

    Attributes:
        model_type (str): The type of model to be used.
        
            Defaults to ``'GP'``. Options are as follows:
        
            - ``'GP'``: Gaussian Process with default settings (Matern Kernel, Gamma covariance priors)
            - ``'MixedGP'``: GP with mixed parameter types (continuous, categorical). Will be re-selected
                by default if 'GP' is selected and input space is mixed.
            - ``'DKL'``: GP with a NN feature-extractor (deep kernel learning)
            - ``'GPflat'``: GP without priors. May result in optimization instability, but removes bias
                for special situations.
            - ``'GPprior'``: GP with custom priors on the mean, likelihood, and covariance
            - ``'MTGP'``: Multi-task GP for multi-output optimization. Will be re-selected by default
                if 'GP' is selected and the input space contains Task parameters.
            - ``'DNN'``: Dropout neural network. Uses MC sampling to mask neurons during training and
                to estimate uncertainty.
              
        hps (dict): Optional surrogate function hyperparameters.
        mll (ExactMarginalLogLikelihood): The marginal log likelihood of the model.
        torch_model (torch.nn.Module): The torch model for the surrogate.
        loss (float): The loss of the model.
        r2_score (float): The R2 score of the model.
    """
    def __init__(self,
                 model_type: str = 'GP',
                 seed: int | None = None,
                 verbose: bool = False,
                 hps: dict = {}):
        
        super().__init__(model_type=model_type, seed=seed, verbose=verbose)
        
        # Optional surrogate function hyperparameters
        self.hps = hps
                
        return
    
    def init_model(self,
                   X: pd.DataFrame,
                   y: pd.Series,
                   cat_dims: list[int],
                   task_feature: int):
        """
        Instantiates the torch model for the surrogate.
        Cannot be called during __init__ normally as X,y are required
        and may not be available until fit methods are called

        Args:
            X (pd.DataFrame): Input parameters for the training data.
            y (pd.Series): Training data responses.
            cat_dims (list): A list of indices for categorical dimensions in the input data.

        Returns:
            None. Updates surrogate attributes, including self.torch_model.
        
        Raises:
            TypeError: If cat_dims is not a list of integers.
            
        """

        # Once the model is created, steward the order of data columns
        self.X_order = X.columns.tolist()
        self.y_name = y.name

        X_p, y_p = self._prepare(X, y)
    
        if not isinstance(cat_dims, list):
            raise TypeError('cat_dims must be a list')
        if not all([isinstance(c, int) for c in cat_dims]):
            raise TypeError('cat_dims must be a list of integers')

        if issubclass(model_class_dict[self.model_type], GPyTorchModel):
            if self.model_type == 'GP' and cat_dims:  # If cat_dims is not an empty list, returns True
                self.torch_model = model_class_dict['MixedGP'](train_X=X_p, train_Y=y_p, cat_dims=cat_dims)
            else:
                if self.model_type == 'MTGP':
                    self.torch_model = model_class_dict[self.model_type](
                        train_X=X_p, train_Y=y_p, task_feature=task_feature, **self.hps)
                else:
                    # Note: Doesn't matter if input empty dictionary as self.hps for model without those additional args
                    self.torch_model = model_class_dict[self.model_type](train_X=X_p, train_Y=y_p, **self.hps)
        else:
            self.torch_model = model_class_dict[self.model_type](train_X=X_p, train_Y=y_p, **self.hps).to(TORCH_DTYPE)

        return

    def fit(self,
            X: pd.DataFrame,
            y: pd.Series,
            cat_dims=None,
            task_feature=None):
        """
        Fits the surrogate model to data

        Args:
            X (pd.DataFrame): Input parameters for the training data
            y (pd.Series): Training data responses
            cat_dims (list, optional): A list of indices for categorical dimensions in the input data. Default is ``None``.

        Returns:
            None. Updates the surrogate model attributes, including regressed parameters.
        """
     
        # Instantiate self.torch_model
        self.init_model(X, y, cat_dims, task_feature)
        X_p, y_p = self._prepare(X, y)

        # Save the raw data inputs into the model metadata
        self.train_X = X
        self.train_Y = y
        self.cat_dims = cat_dims
        self.task_feature = task_feature
        
        # Train
        if self.verbose:
            print('Fitting surrogate model [...]')
        
        if isinstance(self.torch_model, GPyTorchModel):
            self.loss_fcn = ExactMarginalLogLikelihood(self.torch_model.likelihood, self.torch_model)
            if self.model_type == 'DKL':
                optimizer = fit_gpytorch_mll_torch
            else:
                optimizer = fit_gpytorch_mll_scipy

            try:
                fit_gpytorch_mll(self.loss_fcn, optimizer=optimizer)
            except Exception:
                try:
                    fit_gpytorch_mll(self.loss_fcn, optimizer=fit_gpytorch_mll_torch)
                except Exception:
                    raise SurrogateFitError('BoTorch model failed to fit')
        else:
            self.loss_fcn = nn.MSELoss()
            fit_pytorch(self.torch_model, X_p, y_p, loss_fcn=self.loss_fcn, verbose=self.verbose)

        self.is_fit = True

        # Quick evaluation
        self.score(X, y)
        
        return
    
    def score(self,
              X: pd.DataFrame,
              y: pd.Series) -> tuple:
        """
        Computes simple model statistics on a dataset.

        Args:
            X (pd.DataFrame): Input parameters for the evaluation data.
            y (pd.DataFrame): Evaluation data responses.

        Returns:
            tuple: A tuple containing the loss and R2 score of the evaluation.
        """

        X_p, y_p = self._prepare(X, y)
        
        if self.model_type == 'DKL':
            self.torch_model(X_p)
        
        # Quick evaluation
        # Suppress an error that will be generated when evaluating R2_training
        warnings.filterwarnings('ignore', message=r'[.\n]*The input matches the stored training data*')
        mu_pred, _ = self.predict(X)
        corr_matrix = np.corrcoef(mu_pred.detach().cpu().flatten(),
                                  y_p.detach().cpu().flatten())
        
        # Calculate a final loss and R2 train score
        loss = self.loss = self.loss_fcn(self.torch_model(X_p), y_p).sum().detach().cpu().data.numpy().tolist()
        score = self.r2_score = corr_matrix[0][1]**2
        
        return loss, score
    
    def predict(self,
                X: pd.DataFrame,
                q: float | None = None):
        """
        Computes model predictions over new experimental space.

        Args:
            X (pd.DataFrame): Input parameters for the prediction space.
            q (float, op)

        Returns:
            tuple: A tuple containing:
                - mu (ndarray): Mean responses for each experiment in the prediction space.
                - q_pred | sd (ndarray): Quantile or standard deviation of the predicted response for each experiment.
        """
        
        X_p = self._prepare(X)
        
        pred_posterior = self.torch_model.posterior(X_p)

        # We would prefer to have stability in the mean of ensemble models,
        # So, we will not re-sample for prediction but use forward methods
        if isinstance(self.torch_model, EnsembleModel):
            mu = self.torch_model.forward(X_p).detach()
        else:
            mu = pred_posterior.mean.detach().cpu().squeeze(-1)

        if q is not None:
            if (q < 0) or (q > 1):
                raise ValueError('Quantile must be between 0 and 1')
            q_pred = pred_posterior.quantile(torch.tensor(q)).detach().cpu().squeeze(-1)
            return mu, q_pred
        else:
            sd = pred_posterior.variance.detach().cpu().squeeze(-1)**0.5
            return mu, sd

    def save_state(self) -> dict:
        """
        Saves the state of the SurrogateBoTorch model.

        Returns:
            dict: A dictionary containing the state of the SurrogateBoTorch model.
        """
        obj_dict = {'model_attrs': {}}

        model_attrs = ['model_type', 'seed', 'hps',
                       'is_fit',
                       'cat_dims', 'task_feature',
                       'X_order', 'y_name']
        for attr in model_attrs:
            obj_dict['model_attrs'][attr] = getattr(self, attr)
 
        obj_dict['train_X'] = self.train_X.to_dict()
        obj_dict['train_Y'] = self.train_Y.to_dict()

        # Use torch's built in state_dict function
        obj_dict['torch_params'] = tensordict_to_dict(self.torch_model.state_dict())

        return obj_dict
    
    @classmethod
    def load_state(cls,
                   obj_dict: dict):
        """
        Generates a BoTorch surrogate model from a state dictionary (saved after fitting).
        Note that the training data is necessary to instantiate a matching model,
        but no fitting occurs here.

        Args:
            obj_dict (OrderedDict): The state dictionary of a previously fit BoTorch model.

        Returns:
            None. Updates the parameters of the model.
        """

        new_model = cls(model_type=obj_dict['model_attrs']['model_type'],
                        seed=obj_dict['model_attrs']['seed'], verbose=False,
                        hps=obj_dict['model_attrs']['hps'])

        # Directly unpack all of the entries in model_attrs
        for k, v in obj_dict['model_attrs'].items():
            setattr(new_model, k, v)

        # Load data objects
        new_model.train_X = pd.DataFrame(obj_dict['train_X'], columns=new_model.X_order)
        new_model.train_Y = pd.Series(obj_dict['train_Y'], name=new_model.y_name)
 
        # Re-initialize torch model
        new_model.init_model(new_model.train_X, new_model.train_Y, new_model.cat_dims, new_model.task_feature)

        # Load saved parameters into the torch model
        new_model.torch_model.load_state_dict(dict_to_tensordict(obj_dict['torch_params']))

        # Rebuild MLL which would have been done during fitting
        new_model.torch_model.training = False
        if isinstance(new_model.torch_model, GPyTorchModel):
            new_model.loss_fcn = ExactMarginalLogLikelihood(new_model.torch_model.likelihood, new_model.torch_model)
        else:
            new_model.loss_fcn = nn.MSELoss()
        new_model.score(new_model.train_X, new_model.train_Y)
        
        return new_model



"""PyTests for obsidian.experiment"""

from obsidian.tests.param_configs import test_X_space

from obsidian.experiment import ExpDesigner, Simulator
from obsidian.experiment.benchmark import rosenbrock, shifted_parab, ackley

import numpy as np
import pytest

test_methods = ['DOE_full', 'DOE_res4', 'Sobol', 'Random', 'Custom', 'LHS']


@pytest.mark.fast
@pytest.mark.parametrize('X_space', test_X_space)
@pytest.mark.parametrize('method', test_methods)
def test_designer(X_space, method):
    designer = ExpDesigner(X_space, seed=0)
    m = 20
    custom_X = np.random.uniform(0, 1, size=(m, X_space.n_dim))
    if method == 'Custom':
        X0 = designer.initialize(m_initial=m, method=method, sample_custom=custom_X)
    elif method in ['DOE_full', 'DOE_res4'] and X_space.X_discrete:
        pass
    else:
        X0 = designer.initialize(m_initial=m, method=method)
        
    if method == 'LHS':
        X0 = designer.initialize(m_initial=X_space.n_dim*10, method=method)
        for param in X_space:
            if param in X_space.X_discrete:
                assert np.all([level in X0[param.name].unique() for level in param.categories])
    designer.__repr__()


test_functions = [shifted_parab, rosenbrock, ackley]


@pytest.fixture(params=test_X_space)
def X_s(request):
    return request.param


@pytest.fixture
def X0(X_s):
    designer = ExpDesigner(X_s, seed=0)
    X0 = designer.initialize(m_initial=20, method='LHS')
    return X0


@pytest.mark.fast
@pytest.mark.parametrize('function', test_functions)
def test_simulator(function, X0, X_s):
    simulator = Simulator(X_s, function)
    simulator.simulate(X0)
    simulator.__repr__()


if __name__ == '__main__':
    pytest.main([__file__, '-m', 'fast'])



"""Custom objective function formulations"""

from .base import Objective

from obsidian.parameters import ParamSpace, Param_Continuous, Target
from obsidian.utils import tensordict_to_dict
from obsidian.config import TORCH_DTYPE

import torch
from torch import Tensor


class Identity_Objective(Objective):
    """
    Dummy multi-output objective function that simply returns the input Y.
    """
    def __init__(self,
                 mo: bool = False) -> None:
        super().__init__(mo)
    
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        return self.output_shape(samples)


class Product_Objective(Objective):
    """
    Objective function that computes the weighted products of other objectives
    
    Args:
        ind (tuple[int]): The indices of objectives to be used in a product
        weights (list[float | int], optional): The weights corresponding to
            indexed objectives. Defaults to ``(1,)*len(ind)``.
        const (float | int): A constant value that can be added to the product
        new_dim (bool, optional): Whether to create a new objective dimension
            from this product. Default is ``False``. Setting to ``True`` will
            make this the only output objective.
        
    """
    def __init__(self,
                 ind: tuple[int],
                 weights: list[float | int] | None = None,
                 const: float | int = 0,
                 new_dim: bool = True) -> None:
        super().__init__(new_dim)
        # Always MOO if dim is being added, always SOO otherwise
        if weights is None:
            weights = [1]*len(ind)
        if len(weights) != len(ind):
            raise ValueError('The length of weights and indices must be the same')
        self.register_buffer('ind', torch.tensor(ind, dtype=torch.int))
        self.register_buffer('weights', torch.tensor(weights, dtype=TORCH_DTYPE))
        self.register_buffer('const', torch.tensor(const, dtype=TORCH_DTYPE))
        self.register_buffer('new_dim', torch.tensor(new_dim, dtype=torch.bool))
    
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (ind={self.ind.tolist()}, \
            weights={self.weights.tolist()}, const={self.const})'
    
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        obj = (self.weights*samples[..., self.ind]).prod(dim=-1, keepdim=True) + self.const
        if self.new_dim:
            return torch.concat((samples, obj), dim=-1)
        else:
            return self.output_shape(obj)


class Divide_Objective(Objective):
    """
    Objective function that computes the weighted quotient of other objectives
    
    (w_num * samples[..., ind_num])/(w_denom * samples[..., ind_denom]) + const
    
    Args:
        ind_num (int): The index of the objective to be used in the numerator
        w_num (float | int, optional): The weight corresponding to numerator
            objective. Defaults to ``1``.
        ind_denom (int): The index of the objective to be used in the denominator
        w_denom (float | int, optional): The weight corresponding to denominator
            objective. Defaults to ``1``.
        const (float | int): A constant value that can be added to the quotient
        new_dim (bool, optional): Whether to create a new objective dimension
            from this quotient. Default is ``False``. Setting to ``True`` will
            make this the only output objective.
        
    """
    def __init__(self,
                 ind_num: int,
                 ind_denom: int,
                 w_num: float | int = 1,
                 w_denom: float | int = 1,
                 const: float | int = 0,
                 new_dim: bool = True) -> None:
        super().__init__(new_dim)
        # Always MOO if dim is being added, always SOO otherwise

        self.register_buffer('ind_num', torch.tensor(ind_num, dtype=torch.int))
        self.register_buffer('w_num', torch.tensor(w_num, dtype=TORCH_DTYPE))
        self.register_buffer('ind_denom', torch.tensor(ind_denom, dtype=torch.int))
        self.register_buffer('w_denom', torch.tensor(w_denom, dtype=TORCH_DTYPE))
        self.register_buffer('const', torch.tensor(const, dtype=TORCH_DTYPE))
        self.register_buffer('new_dim', torch.tensor(new_dim, dtype=torch.bool))
        
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (num={self.w_num} * {self.ind_num}, \
            denom={self.w_denom} * {self.ind_denom}, const={self.const})'
            
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        w_num = (self.w_num*samples[..., self.ind_num]).unsqueeze(-1)
        w_denom = (self.w_denom*samples[..., self.ind_denom]).unsqueeze(-1)
        obj = w_num/w_denom + self.const
        
        if self.new_dim:
            return torch.concat((samples, obj), dim=-1)
        else:
            return self.output_shape(obj)


class Feature_Objective(Objective):
    """
    Creates an objective function which creates a new outcome as a linear combination of input features.
    
    Always a multi-output objective.

    Args:
        X_space (ParamSpace): The parameter space.
        ind (tuple[int]): The indices of the parameters in the real space
            to be used as features.
        coeff (list[float | int], optional): The coefficients corresponding to each feature.
            Defaults to ``[0]``.

    Raises:
        ValueError: If the length of `ind` and `coeff` are not the same.
        TypeError: If the indices for input objectives are not for continuous parameters.
    """
    
    def __init__(self,
                 X_space: ParamSpace,
                 ind: tuple[int],
                 coeff: list[float | int]) -> None:
        
        super().__init__(mo=True)
        
        if len(ind) != len(coeff):
            raise ValueError("featureids and coeff must have the same length")
        
        for i in ind:
            if not isinstance(X_space.params[i], Param_Continuous):
                raise TypeError('Indices for input objectives must be for continuous parameters only')
        
        self.register_buffer('coeff', torch.tensor(coeff, dtype=TORCH_DTYPE))
        self.register_buffer('ind', torch.tensor(ind, dtype=torch.long))
        self.X_space = X_space
    
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (ind={self.ind.tolist()}, coeff={self.coeff.tolist()})'

    def forward(self,
                samples: Tensor,
                X: Tensor) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        # Ydim = s * b * q * m
        # Xdim = b * q * d
        # if q is 1, it is omitted

        X_u_all = []  # Create unscaled X
        for i in self.ind:
            X_u_all.append(torch.tensor(
                self.X_space[i].decode(X[..., i].detach().numpy()),
                dtype=TORCH_DTYPE).unsqueeze(-1))  # Add output dimension
            
        X_u = torch.concat(X_u_all, dim=-1).unsqueeze(0)  # Add sample dimension to X_u
        cX = self.coeff * X_u
        
        # Make sure that the new obejctive matches Y in shape
        # Sum the C*X features as a new output
        feature_obj = cX.sum(dim=-1, keepdim=True).repeat_interleave(samples.shape[0], 0)
        
        total_obj = torch.concat([samples, feature_obj], dim=-1)
        
        return total_obj
    
    def save_state(self) -> dict:
        """Saves the objective to a state dictionary"""
        obj_dict = {'name': self.__class__.__name__,
                    'state_dict': tensordict_to_dict(self.state_dict()),
                    'X_space': self.X_space.save_state()}
        
        return obj_dict
    
    @classmethod
    def load_state(cls, obj_dict: dict):
        """Loads the objective from a state dictionary"""
        new_obj = cls(ParamSpace.load_state(obj_dict['X_space']),
                      **obj_dict['state_dict'])
        
        return new_obj


class Utopian_Distance(Objective):
    """
    A custom objective for calculating the distance between an output response
    and a most-desirable, ideal/utopian point.

    Args:
        utopian (tuple[float | int]): A list of values representing the utopian point.
        targets (Target | list[Target]): A single Target object or a list of Target objects.

    Attributes:
        u_t (Tensor): The utopian coordinates in transformed space
    
    Raises:
        TypeError: If the input is not a list of Target objects.
        ValueError: If the length of the utopian point and targets are not the same.

    """

    def __init__(self,
                 utopian: list[float | int],
                 targets: Target | list[Target]) -> None:

        if not isinstance(targets, list):
            targets = [targets]
        if not all(isinstance(t, Target) for t in targets):
            raise TypeError('All elements in the list must be Target objects')

        if len(utopian) != len(targets):
            raise ValueError("Length of utopian and targets must be the same")

        super().__init__(mo=len(targets) > 1)
        
        self.register_buffer('utopian', torch.tensor(utopian, dtype=TORCH_DTYPE))
        self.targets = targets
        
        # Must transform the utopian coordinates into the transformed space where
        # the optimizer operates on target samples
        u_t = [t.transform_f(u) for u, t in zip(utopian, targets)]
        self.u_t = torch.tensor(u_t, dtype=TORCH_DTYPE).flatten()

    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        distance = (-1)*(self.u_t-samples).abs()
        
        return self.output_shape(distance)
    
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (utopian={self.utopian.tolist()})'
    
    def save_state(self) -> dict:
        """Saves the objective to a state dictionary"""
        obj_dict = {'name': self.__class__.__name__,
                    'state_dict': tensordict_to_dict(self.state_dict()),
                    'targets': [t.save_state() for t in self.targets]}
        
        return obj_dict
    
    @classmethod
    def load_state(cls, obj_dict: dict):
        """Loads the objective from a state dictionary"""
        new_obj = cls(targets=[Target.load_state(t_dict) for t_dict in obj_dict['targets']],
                      **obj_dict['state_dict'])
        
        return new_obj


class Bounded_Target(Objective):
    """
    Represents a bounded target objective for multi-output and single-output
    acquisition objectives

    Args:
        bounds (list[tuple[float, float] | None]): The bounds for each target.
            If a bound is None, it is ignored.
        targets (Target | list[Target]): The target or list of targets.
        tau (float, optional): The temperature parameter for the sigmoid function.
            Defaults to ``1e-3``.

    Attributes:
        lb (torch.Tensor): The lower bounds for the targets.
        ub (torch.Tensor): The upper bounds for the targets.
        tau (torch.Tensor): The temperature parameter for the sigmoid function.
        ind (torch.Tensor): The indices of the non-None bounds.

    """

    def __init__(self,
                 bounds: list[tuple[float, float] | None],
                 targets: Target | list[Target],
                 tau: float = 1e-3) -> None:
        if not isinstance(targets, list):
            targets = [targets]

        if not all(isinstance(t, Target) for t in targets):
            raise TypeError('All elements in the list must be Target objects')

        self.targets = targets

        if len(bounds) != len(targets):
            raise ValueError("Length of bounds and targets must be the same")

        super().__init__(mo=len(targets) > 1)

        # Only store and process the indicated bounds
        ind = []
        lb_t = []
        ub_t = []
        for i, (b, target) in enumerate(zip(bounds, targets)):
            if b is not None:
                lb_t.append(target.transform_f(b[0]).iloc[0])
                ub_t.append(target.transform_f(b[1]).iloc[0])
                ind.append(i)
    
        self.bounds = bounds
        self.ind = torch.tensor(ind)
        self.lb = torch.tensor(lb_t, dtype=TORCH_DTYPE).flatten()
        self.ub = torch.tensor(ub_t, dtype=TORCH_DTYPE).flatten()
        tau = torch.tensor(tau, dtype=TORCH_DTYPE)

        self.register_buffer('tau', tau)

    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (ind={self.ind.tolist()}, bounds={self.bounds})'

    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        approx_lb = torch.sigmoid((samples[..., self.ind] - self.lb) / self.tau)
        approx_ub = torch.sigmoid((self.ub - samples[..., self.ind]) / self.tau)
        product = approx_lb * approx_ub
        out = samples.to(TORCH_DTYPE)
        out[..., self.ind] = product
        return self.output_shape(out)

    def save_state(self) -> dict:
        """Saves the objective to a state dictionary"""
        obj_dict = {'name': self.__class__.__name__,
                    'state_dict': tensordict_to_dict(self.state_dict()),
                    'bounds': self.bounds,
                    'targets': [t.save_state() for t in self.targets]}
        
        return obj_dict
    
    @classmethod
    def load_state(cls, obj_dict: dict):
        """Loads the objective from a state dictionary"""
        new_obj = cls(targets=[Target.load_state(t_dict) for t_dict in obj_dict['targets']],
                      bounds=obj_dict['bounds'],
                      **obj_dict['state_dict'])
        
        return new_obj


class Index_Objective(Objective):
    """
    Creates an objective function that returns the single-objective output
    from a multi-output model, at the specified index.
    
    Single-objective when index is int; multi-objective if tuple.

    Args:
        ind (int | tuple[int]): The index of the value to be returned.
    """
    def __init__(self,
                 ind: int | tuple[int] = 0) -> None:
        super().__init__(mo=not isinstance(ind, int))
        self.register_buffer('ind', torch.tensor(ind))
        
    def __repr__(self):
        """String representation of object"""
        return f'{self.__class__.__name__} (ind={self.ind.item()})'
        
    def forward(self,
                samples: Tensor,
                X: Tensor | None = None) -> Tensor:
        """
        Evaluate the objective function on the candidate set samples, X
        """
        return samples[..., self.ind]



"""PyTests for obsidian.optimizer under multi-output usage"""

from obsidian.tests.param_configs import X_sp_cont_ndims

from obsidian.parameters import Target
from obsidian.experiment import ExpDesigner, Simulator
from obsidian.optimizer import BayesianOptimizer
from obsidian.experiment.benchmark import two_leaves
from obsidian.tests.utils import equal_state_dicts

from botorch.models.ensemble import EnsembleModel
      
import pandas as pd
import numpy as np
import pytest


@pytest.fixture()
def X_space():
    return X_sp_cont_ndims[2]


@pytest.fixture()
def Z0(X_space):
    designer = ExpDesigner(X_space, seed=0)
    X0 = designer.initialize(m_initial=len(X_space)*2, method='LHS')
    simulator = Simulator(X_space, two_leaves, eps=0.05)
    y0 = simulator.simulate(X0)
    Z0 = pd.concat([X0, y0], axis=1)
    return Z0


target = [
    Target(name='Response 1', f_transform='Standard', aim='max'),
    Target(name='Response 2', f_transform='Standard', aim='max')
]


@pytest.mark.parametrize('surrogate', [pytest.param('GP', marks=pytest.mark.fast),
                                       'GPflat',
                                       'GPprior',
                                       pytest.param('DKL', marks=pytest.mark.slow),
                                       'DNN'])
def test_optimizer_fit(X_space, surrogate, Z0, serial_test=True):
    optimizer = BayesianOptimizer(X_space, surrogate=surrogate, seed=0, verbose=3)
    
    tol = 1e-2 if surrogate == 'DNN' else 1e-5
    
    optimizer.fit(Z0, target=target)
    if serial_test:
        obj_dict = optimizer.save_state()
        optimizer_2 = BayesianOptimizer.load_state(obj_dict)
        obj_dict2 = optimizer_2.save_state()
        assert equal_state_dicts(obj_dict, obj_dict2)
        optimizer_2.__repr__()

        y_pred = optimizer.predict(optimizer.X_train)
        y_pred_2 = optimizer_2.predict(optimizer.X_train)
        y_error = ((y_pred_2-y_pred)/y_pred.max(axis=0))

        # Only check the mean for ensemble models, as the lb/ub are not deterministic
        if isinstance(optimizer.surrogate[0].torch_model, EnsembleModel):
            check_cols = [col for col in y_pred if '(pred)' in col]
            y_error = y_error[check_cols]
        
        assert abs(y_error.values).max() < tol, 'Prediction error in loading parameters of saved optimizer'


# Generate a baseline optimizer to use for future tests
base_X_space = X_sp_cont_ndims[2]
optimizer = BayesianOptimizer(base_X_space, surrogate='GP', seed=0, verbose=0)
designer = ExpDesigner(base_X_space, seed=0)
X0 = designer.initialize(m_initial=6, method='LHS')
simulator = Simulator(base_X_space, two_leaves, eps=0.05)
y0 = simulator.simulate(X0)
Z0_base = pd.concat([X0, y0], axis=1)
optimizer.fit(Z0_base, target=target)

# Run very short optimizations for testing
test_config = {'optim_samples': 2, 'optim_restarts': 2}


def test_fit_nan():
    Z0_sample = Z0_base.copy()
    for col in Z0_sample.columns:
        Z0_sample.loc[Z0_sample.sample(frac=0.1).index, col] = np.nan
    optimizer_nan = BayesianOptimizer(base_X_space, surrogate='GP', seed=0, verbose=0)
    optimizer_nan.fit(Z0_sample, target=target)


@pytest.mark.fast
def test_optimizer_pending():
    X_suggest, eval_suggest = optimizer.suggest(m_batch=2, **test_config)
    X_suggest, eval_suggest = optimizer.suggest(m_batch=1, **test_config, X_pending=X_suggest)
    X_suggest, eval_suggest = optimizer.suggest(m_batch=1, **test_config, X_pending=X_suggest, eval_pending=eval_suggest)


@pytest.mark.parametrize('m_batch', [pytest.param(1, marks=pytest.mark.fast), 3])
@pytest.mark.parametrize('fixed_var', [None, {'Parameter 1': 5}])
def test_optimizer_suggest(m_batch, fixed_var):
    X_suggest, eval_suggest = optimizer.suggest(m_batch=m_batch, fixed_var=fixed_var,
                                                acquisition=['NEHVI', 'SF'], **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)


def test_suggest_searchspace():
    optimizer.X_space[0].set_search(2, 8)
    
    X_suggest, eval_suggest = optimizer.suggest(m_batch=2, **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
    
    optimizer.X_space.open_search()
    

test_aqs = ['NEHVI',
            {'NEHVI': {'ref_point': [0.1, 0.1]}},
            'EHVI',
            {'EHVI': {'ref_point': [0.1, 0.1]}},
            'NParEGO',
            {'NParEGO': {'scalarization_weights': [5, 1]}},
            'Mean',
            'SF',
            'RS',
            ]


@pytest.mark.parametrize('aq', test_aqs)
def test_optimizer_aqs(aq):
    X_suggest, eval_suggest = optimizer.suggest(m_batch=2,  acquisition=[aq], **test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)
   

@pytest.mark.fast
def test_optimizer_maximize():
    X_suggest, eval_suggest = optimizer.maximize(**test_config)
    df_suggest = pd.concat([X_suggest, eval_suggest], axis=1)


if __name__ == '__main__':
    pytest.main([__file__, '-m', 'not slow'])



from .utils import add_tab, make_input, make_dropdown, make_switch, make_slider, make_knob, make_table, make_collapse
from .utils import center
import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH
import pandas as pd
import base64
import io
from dash.exceptions import PreventUpdate
from obsidian.parameters import ParamSpace, Param_Categorical, Param_Ordinal, Param_Continuous, Target, Parameter
from pandas.api.types import is_numeric_dtype


def setup_data(app, app_tabs, default_data, default_Xspace):
    
    # Data upload
    uploader = dcc.Upload(id='uploader-X0',
                          children=html.Div(['Upload Data: Drag and Drop or ',
                                             html.A('Select Files')]),
                          style={
                              'width': '100%', 'height': '60px',
                              'lineHeight': '60px', 'borderWidth': '1px',
                              'borderStyle': 'dashed', 'borderRadius': '5px',
                              'textAlign': 'center', 'margin': '10px'
                              },
                          multiple=False,
                          filename='Example Data'
                          )
    
    # Example data download
    template_downloader = html.Div(children=[dbc.Button('Download Template Data', id='button-download_data',
                                                        className='me-2', color='primary'),
                                             dcc.Download(id='downloader-X0_template')],
                                   style={'textAlign': 'center'})
    
    # Data preview
    preview = html.Div(id='table-X0', children=dbc.Card(
        [dbc.CardHeader('Input Data'),
         dbc.CardBody(make_table(default_data, fill_width=True), id='table-X0-body'),
         dbc.CardFooter(html.Div([
             html.I(id='info-data', className='bi bi-info-circle-fill me-2'),
             dbc.Tooltip('Input data must be a CSV file and must include column headers for the input\
                         parameters and response variable(s), with one row per observation. Download\
                         template data (left) for example.',
                         target='info-data', placement='top', style={'text-transform': 'none'}),
             dbc.FormText('Example Data', color='info', style={'font-size': '1em', 'font-style': 'italic'},
                          id='table-X0-footer')
             ],
             style={'textAlign': 'center'}))
         ]))
    
    preview_uploader = dbc.Row([dbc.Col([uploader, template_downloader], width=4),
                                dbc.Col(preview, width=8)],
                               style={'margin-top': '15px'})
    
    # Data store
    storage_X0 = dcc.Store(id='store-X0', data=default_data.to_dict())
    storage_X0_template = dcc.Store(id='store-X0_template', data=default_data.to_dict())
    storage_Xspace = dcc.Store(id='store-Xspace', data=default_Xspace.save_state())
    
    # Parameter space table
    xspace = html.Div([html.Div(id='div-xspace', children=[])])

    # Selecting the response variable
    ycol = dbc.Row(dbc.Col(dbc.Card([dbc.CardHeader([html.I(id='info-response_col',
                                                            className='bi bi-info-circle-fill me-2'),
                                                     dbc.Tooltip('The response is the variable which is\
                                                                 chosen for optimization (e.g. yield, productivity);\
                                                                 it is the target of the objective function before\
                                                                 including optional transformations.',
                                                                 target='info-response_col', placement='top',
                                                                 style={'text-transform': 'none'}),
                                                     'Response Selection']),
                   dbc.CardBody(html.Div(id='div-response_name',
                                         children=[make_dropdown('Data Column',
                                                                 'Select column for response',
                                                                 default_data.columns,
                                                                 id='input-response_name',
                                                                 kwargs={'value': default_data.columns[-1]})]))
                    ]), width=4), justify='center')
    
    # Extra div for printing outputs, for troubleshooting
    troubleshoot = html.Div(id='debug-print-data')
    
    # Add all of these elements to the app
    elements = [html.Br(), preview_uploader, html.Hr(), ycol, xspace, storage_X0,
                storage_X0_template, storage_Xspace, html.Hr(), troubleshoot]
    add_tab(app_tabs, elements, 'tab-data', 'Data')
    setup_data_callbacks(app)
    
    return


def setup_data_callbacks(app):
    
    # Save the uploaded data into the data-store
    @app.callback(
        Output('store-X0', 'data'),
        Input('uploader-X0', 'contents'),
    )
    def save_X0(contents):
        if contents is None:
            raise PreventUpdate
        content_type, content_string = contents.split(',')
        decoded = base64.b64decode(content_string)
        df = pd.read_csv(io.StringIO(decoded.decode('utf-8')))
        return df.to_dict()
    
    # View the uploaded data in a Data Table
    @app.callback(
        Output('table-X0-body', 'children'),
        Output('table-X0-footer', 'children'),
        Input('store-X0', 'data'),
        State('uploader-X0', 'filename')
    )
    def preview_X0(data, filename):
        df = pd.DataFrame(data)
        return make_table(df, fill_width=True), filename
    
    # Download template data
    @app.callback(
        Output('downloader-X0_template', 'data'),
        Input('button-download_data', 'n_clicks'),
        State('store-X0_template', 'data'),
        prevent_initial_call=True  # otherwise, the download CSV appears right away
    )
    def download_X0_template(n_clicks, data):
        df = pd.DataFrame(data)
        return dcc.send_data_frame(df.to_csv, 'Example_Data.csv', index=False)
    
    # Select a response variable
    @app.callback(
        Output('input-response_name', 'options'),
        Output('input-response_name', 'value'),
        Input('store-X0', 'data')
    )
    def choose_col(data):
        df = pd.DataFrame(data)
        return df.columns, df.columns[-1]
            
    # Select types for each input variable
    @app.callback(
        Output('div-xspace', 'children'),
        Input('store-X0', 'data'),
        Input('input-response_name', 'value'),
        State('store-Xspace', 'data')
    )
    def update_xspace_types(data, ycol, Xspace_save):
        
        X_space = ParamSpace.load_state(Xspace_save)
        
        df_X0 = pd.DataFrame(data)
        xcols = [x for x in df_X0.columns if x != ycol]
        
        param_types = ['Numeric', 'Categorical', 'Ordinal']
        
        cols = []
        for i, x in enumerate(xcols):
            if x in X_space.X_names:
                x_idx = X_space.X_names.index(x)
                param = X_space.params[x_idx]
                if isinstance(param, Param_Continuous):
                    param_type = 'Numeric'
                elif isinstance(param, Param_Categorical):
                    param_type = 'Categorical'
                elif isinstance(param, Param_Ordinal):
                    param_type = 'Ordinal'
            else:
                if not is_numeric_dtype(df_X0[x]):
                    param_type = 'Categorical'
                else:
                    param_type = 'Numeric'
            
            cols.append(dbc.Col(children=dbc.Card([
                dbc.CardHeader(f'{x}'),
                dbc.CardBody([make_dropdown('Type', f'Select parameter type for {x}', param_types,
                                            id={'type': 'input-param_type', 'index': x},
                                            kwargs={'value': param_type}),
                              html.Div(id={'type': 'div-param_vals', 'index': x}, children=[]),
                              dcc.Store(id={'type': 'store-param_save', 'index': x}, data={})],)
                ],
                color='primary', outline=True), width=2))
 
        return dbc.Container(dbc.Row(cols, style={'margin-top': '15px'},
                                     justify='center'),
                             fluid=True)
    
    # Select min, max, and/or categories for each input variable
    @app.callback(
        Output({'type': 'div-param_vals', 'index': MATCH}, 'children'),
        Input({'type': 'input-param_type', 'index': MATCH}, 'value'),
        State({'type': 'input-param_type', 'index': MATCH}, 'id'),
        State('store-X0', 'data'),
        State('input-response_name', 'value')
    )
    def update_xspace_vals(param_type, param_id, data, ycol):
        
        df = pd.DataFrame(data)
        x = param_id['index']
        ser_x = df[x]
        
        # All keys must be present, so Numeric still needs "None" categories, etc.
        if param_type == 'Numeric':
            choices = [
                make_input(f'{x} Min Value', f'Select optimization minimum for {x}', df[x].min(),
                           id={'type': 'input-param_min', 'index': x}),
                make_input(f'{x} Max Value', f'Select optimization maximum for {x}', df[x].max(),
                           id={'type': 'input-param_max', 'index': x}),
                dcc.Store(id={'type': 'store-param_categories', 'index': x}, data=None)
                ]
                           
        elif param_type in ['Categorical', 'Ordinal']:
            choices = [dbc.Card(dbc.CardBody([
                       dbc.InputGroup([dbc.Input(id={'type': 'input-new_category', 'index': x},
                                                 placeholder='New category'),
                                       dbc.Button('Add', id={'type': 'button-category_add', 'index': x},
                                                  className='me-2', color='primary'),
                                       dbc.Button('Clear All', id={'type': 'button-category_delete', 'index': x},
                                                  className='me-2', color='danger')]),
                       html.Hr(),
                       html.Div(id={'type': 'div-param_categories', 'index': x}, children=None,
                                style={'textAlign': 'center'})])),
                       dcc.Store(id={'type': 'store-param_categories', 'index': x},
                                 data=', '.join(list(ser_x.sort_values().astype('str').unique()))),
                       html.Div(id={'type': 'input-param_min', 'index': x}),
                       html.Div(id={'type': 'input-param_max', 'index': x}),
                       ]
        return choices

    @app.callback(
        Output({'type': 'store-param_save', 'index': MATCH}, 'data'),
        Input({'type': 'store-param_save', 'index': MATCH}, 'id'),
        Input({'type': 'input-param_type', 'index': MATCH}, 'value'),
        Input({'type': 'input-param_min', 'index': MATCH}, 'value'),
        Input({'type': 'input-param_max', 'index': MATCH}, 'value'),
        Input({'type': 'store-param_categories', 'index': MATCH}, 'data'),
        prevent_initial_call=True  # It takes a second for these input matches to show up
    )
    def update_param_save(param_id, param_type, param_min, param_max, param_cat):
        
        name = param_id['index']
        
        try:
            if param_type == 'Numeric':
                param = Param_Continuous(name, param_min, param_max)
            elif param_type == 'Categorical':
                param = Param_Categorical(name, param_cat)
            elif param_type == 'Ordinal':
                param = Param_Ordinal(name, param_cat)
        
            single_param = ParamSpace([param])
            return single_param.save_state()
        
        except TypeError:
            return None
    
    # Category management for categorical variables
    @app.callback(
        Output({'type': 'store-param_categories', 'index': MATCH}, 'data'),
        Output({'type': 'input-new_category', 'index': MATCH}, 'value'),
        Output({'type': 'button-category_add', 'index': MATCH}, 'n_clicks'),
        Output({'type': 'button-category_delete', 'index': MATCH}, 'n_clicks'),
        Input({'type': 'button-category_add', 'index': MATCH}, 'n_clicks'),
        Input({'type': 'button-category_delete', 'index': MATCH}, 'n_clicks'),
        State({'type': 'store-param_categories', 'index': MATCH}, 'data'),
        State({'type': 'input-new_category', 'index': MATCH}, 'value'),
        State({'type': 'store-param_categories', 'index': MATCH}, 'id'),
        State('store-X0', 'data'),
        prevent_initial_call=True
    )
    def cat_buttons(add_button, clear_button, current_cats, new_cat, param_id, data):
        
        x = param_id['index']
        df = pd.DataFrame(data)
        ser_x = df[x]
        
        if clear_button:
            return ', '.join(list(ser_x.sort_values().astype('str').unique())), '', 0, 0
        
        if add_button:
            if new_cat != '':
                if current_cats is None:
                    current_cats = new_cat
                else:
                    if new_cat not in current_cats:
                        current_cats += ', '+new_cat
            return current_cats, '', 0, 0
    
    # Preview categories for categorical variables
    @app.callback(
        Output({'type': 'div-param_categories', 'index': MATCH}, 'children'),
        Input({'type': 'store-param_categories', 'index': MATCH}, 'data'),
    )
    def preview_cats(current_cats):
        if current_cats is None:
            return None
        else:
            cat_list = current_cats.split(', ')
            
            preview = []
            for cat in cat_list:
                preview += [html.Div(cat, style={'font-size': '0.8em'})]
            return preview

    @app.callback(
        Output('debug-print-data', 'children'),
        Input('store-X0', 'data'),
    )
    def troubleshoot_config(data):
        return
    
    @app.callback(
        Output('store-Xspace', 'data'),
        Input({'type': 'store-param_save', 'index': ALL}, 'data'),
        prevent_initial_call=True
    )
    def save_Xspace(param_saves):
        
        param_list = []
        for param_input in param_saves:
            if param_input:
                param_i = ParamSpace.load_state(param_input)
                param_list += list(param_i.params)
                
        X_space = ParamSpace(param_list)
        
        return X_space.save_state()
    
    return



from .utils import add_tab, make_input, make_dropdown, make_switch, make_slider, make_knob, make_table, make_collapse
import dash_bootstrap_components as dbc
from dash import dcc, html, Dash, dash_table, callback, Output, Input, State, ALL, MATCH

import pandas as pd

from obsidian.plotting.plotly import surface_plot, factor_plot
from .utils import load_optimizer, center


def setup_plots(app, app_tabs):
    
    surface_plot = dbc.Container(children=[
        html.H5('Surface Plot'),
        html.Br(),
        dbc.Row([
            dbc.Col(make_dropdown('X Axis', 'Select the first parameter for the surface plot',
                                  options=[], id='input-surfaceplot-x')),
            dbc.Col(make_dropdown('Y Axis', 'Select the second parameter for the surface plot',
                                  options=[], id='input-surfaceplot-y'))
            ]),
        dbc.Spinner(html.Div(id='graph-surface'), color='primary')
        ])
    factor_plot = dbc.Container(children=[
        html.H5('1D Factor Effects'),
        html.Br(),
        make_dropdown('Factor', 'Select the parameter for the factor effect plot',
                      options=[], id='input-factorplot-x',),
        dbc.Spinner(html.Div(id='graph-factor'), color='primary')
        ])
    
    # Add all of these elements to the app
    elements = [html.Br(), surface_plot, html.Hr(), factor_plot]
    add_tab(app_tabs, elements, 'tab-plots', 'Explore')
    setup_plots_callbacks(app)
    
    return


def setup_plots_callbacks(app):
    
    @app.callback(
        Output('input-surfaceplot-x', 'options'),
        Output('input-surfaceplot-x', 'value'),
        Input('store-fit', 'data'),
        State('store-X0', 'data'),
        State('input-response_name', 'value')
    )
    def select_x_surface_plot(opt_save, X0, response):
        df_X0 = pd.DataFrame(X0)
        if opt_save is None:
            return [], None
        x_options = df_X0.columns.to_list()
        x_options.remove(response)
        return x_options, x_options[0]
    
    @app.callback(
        Output('input-surfaceplot-y', 'options'),
        Output('input-surfaceplot-y', 'value'),
        Input('store-fit', 'data'),
        Input('input-surfaceplot-x', 'value'),
        State('input-surfaceplot-x', 'options')
    )
    def select_y_surface_plot(opt_save, xval, x_options):
        if opt_save is None:
            return [], None
        y_options = x_options
        y_options.remove(xval)
        return y_options, y_options[0]
    
    @app.callback(
        Output('input-factorplot-x', 'options'),
        Output('input-factorplot-x', 'value'),
        Input('store-fit', 'data'),
        State('store-X0', 'data'),
        State('input-response_name', 'value')
    )
    def select_x_factor_plot(opt_save, X0, response):
        df_X0 = pd.DataFrame(X0)
        if opt_save is None:
            return [], None
        x_options = df_X0.columns.to_list()
        x_options.remove(response)
        return x_options, x_options[0]
    
    @app.callback(
        Output('graph-surface', 'children'),
        Input('store-fit', 'data'),
        Input('input-surfaceplot-x', 'value'),
        Input('input-surfaceplot-y', 'value'),
        State('store-config', 'data')
    )
    def graph_surface_plot(opt_save, p_x, p_y, config):
        if opt_save is None:
            return dbc.Alert('Model must be fit first', color='info')
        
        optimizer = load_optimizer(config, opt_save)
        i_x = list(optimizer.X_space.X_names).index(p_x)
        i_y = list(optimizer.X_space.X_names).index(p_y)
        
        splot = surface_plot(optimizer, [i_x, i_y], plot_data=True)
        splot.update_layout(height=600)
        
        graph = dcc.Graph(figure=splot)
        
        return center(graph)
 
    @app.callback(
        Output('graph-factor', 'children'),
        Input('store-fit', 'data'),
        Input('input-factorplot-x', 'value'),
        State('store-config', 'data')
    )
    def graph_factor_plot(opt_save, p_x, config):
        if opt_save is None:
            return dbc.Alert('Model must be fit first', color='info')
        
        optimizer = load_optimizer(config, opt_save)
        i_x = list(optimizer.X_space.X_names).index(p_x)
        
        fplot = factor_plot(optimizer, feature_id=i_x)
        fplot.update_layout(height=600, width=800)
        
        graph = dcc.Graph(figure=fplot)
        
        return center(graph)
 
    return



"""PyTests for obsidian.campaign"""

from obsidian.parameters import Target
from obsidian.experiment import Simulator
from obsidian.experiment.benchmark import two_leaves, shifted_parab
from obsidian.campaign import Campaign, Explainer, calc_ofat_ranges
from obsidian.objectives import Identity_Objective
from obsidian.plotting import plot_interactions, plot_ofat_ranges
from obsidian.exceptions import IncompatibleObjectiveError, UnfitError

from obsidian.tests.param_configs import X_sp_cont_ndims, X_sp_default
from obsidian.tests.utils import DEFAULT_MOO_PATH, DEFAULT_SOO_PATH, equal_state_dicts

import pandas as pd
import pytest
import json
 
# Avoid using TkAgg which causes Tcl issues during testing
import matplotlib
matplotlib.use('inline')

target_test = [
    [Target(name='Response 1', f_transform='Standard', aim='max'),
     Target(name='Response 2', f_transform='Standard', aim='max')],
    Target(name='Response', f_transform='Standard', aim='max')
]


@pytest.mark.parametrize('X_space, sim_fcn, target',
                         [(X_sp_cont_ndims[2], two_leaves, target_test[0]),
                          (X_sp_default, shifted_parab, target_test[1])])
def test_campaign_basics(X_space, sim_fcn, target):
    # Standard usage
    campaign = Campaign(X_space, target)
    simulator = Simulator(X_space, sim_fcn, eps=0.05)
    X0 = campaign.suggest()
    y0 = simulator.simulate(X0)
    Z0 = pd.concat([X0, y0], axis=1)
    
    # Set an objective, suggest, clear
    campaign.set_objective(Identity_Objective(mo=len(campaign.target) > 1))
    campaign.suggest()
    campaign.clear_objective()
    
    # Add, fit, clear, examine
    campaign.add_data(Z0)
    campaign.fit()
    campaign.clear_data()
    campaign.y
    campaign.__repr__()
    
    # Add with iteration, examine, fit, analyze
    Z0['Iteration'] = 5
    campaign.add_data(Z0)
    campaign.y
    campaign.fit()
    campaign.response_max
    campaign.X_best

    # Serialize, deserialize, re-serialize
    obj_dict = campaign.save_state()
    campaign2 = Campaign.load_state(obj_dict)
    obj_dict2 = campaign2.save_state()
    assert equal_state_dicts(obj_dict, obj_dict2), 'Error during serialization'
    
    
# Load default
with open(DEFAULT_MOO_PATH) as json_file:
    obj_dict = json.load(json_file)
campaign = Campaign.load_state(obj_dict)
X_space = campaign.X_space
target = campaign.target


def test_explain():
    # SOO includes discrete variables
    with open(DEFAULT_SOO_PATH) as json_file:
        obj_dict = json.load(json_file)
    campaign = Campaign.load_state(obj_dict)
    
    # Standard usage
    exp = Explainer(campaign.optimizer)
    exp.shap_explain(n=50)
    exp.__repr__

    # Test SHAP plots
    exp.shap_summary()
    exp.shap_summary_bar()
    
    # Test PDP-ICE, with options and  discrete variables
    exp.shap_pdp_ice(ind=3, ice_color_var=None, npoints=10)
    exp.shap_pdp_ice(ind=0, ice_color_var=3, npoints=None)
    exp.shap_pdp_ice(ind=3, hist=True)
    exp.shap_pdp_ice(ind=(0, 3), npoints=5)
    exp.shap_pdp_ice(ind=(3, 0), npoints=None)

    # Test pairwise SHAP analysis, with options
    X_new = campaign.X.iloc[0, :]
    X_ref = campaign.X.loc[1, :]
    df_shap_value_new, fig_bar, fig_line = exp.shap_single_point(X_new)
    df_shap_value_new, fig_bar, fig_line = exp.shap_single_point(X_new, X_ref=X_ref)

    # Test sensitivity analysis, with options
    df_sens = exp.sensitivity()
    df_sens = exp.sensitivity(X_ref=X_ref)


X_ref_test = [None,
              campaign.X.iloc[campaign.y.idxmax()['Response 1'], :]]


@pytest.mark.parametrize('X_ref', X_ref_test)
def test_analysis(X_ref):
    # OFAT ranges with/out interactions and with/out X_ref
    ofat_ranges, _ = calc_ofat_ranges(campaign.optimizer, threshold=0.5, X_ref=X_ref, calc_interacts=False)
    ofat_ranges, cor = calc_ofat_ranges(campaign.optimizer, threshold=0.5, X_ref=X_ref)
    plot_interactions(campaign.optimizer, cor)
    plot_ofat_ranges(campaign.optimizer, ofat_ranges)
    
    # OFAT ranges where all results should be NaN
    ofat_ranges, cor = calc_ofat_ranges(campaign.optimizer, threshold=9999, X_ref=X_ref)
    plot_interactions(campaign.optimizer, cor)
    plot_ofat_ranges(campaign.optimizer, ofat_ranges)


# VALIDATION TESTS - Force errors to be raised in object usage

@pytest.mark.fast
def test_campaign_validation():
    
    # Missing X names
    random_data = pd.DataFrame(data={'A': [1, 2, 3], 'B': [4, 5, 6]})
    with pytest.raises(KeyError):
        campaign.add_data(random_data)
    
    # Missing Y names
    with pytest.raises(KeyError):
        campaign.add_data(campaign.X)

    # Missing data
    with pytest.raises(ValueError):
        campaign2 = Campaign(X_space, target)
        campaign2.fit()

    # Bad objective
    with pytest.raises(IncompatibleObjectiveError):
        campaign.set_objective(Identity_Objective(mo=False))


@pytest.mark.fast
def test_explainer_validation():
    
    # Unfit optimizer
    campaign2 = Campaign(X_space, target)
    with pytest.raises(UnfitError):
        exp = Explainer(campaign2.optimizer)
    
    # Unfit SHAP
    exp = Explainer(campaign.optimizer)
    with pytest.raises(UnfitError):
        exp.shap_summary()
    
    # Unfit SHAP
    with pytest.raises(UnfitError):
        exp.shap_summary_bar()
    
    # Unfit SHAP
    with pytest.raises(UnfitError):
        exp.shap_single_point(X_new=campaign.optimizer.X_best_f)
    
    random_data = pd.DataFrame(data={'A': [1], 'B': [4]})
    long_data = pd.DataFrame(data={'Parameter 1': [1, 2], 'Parameter 2': [1, 2]})
    
    # Missing X names
    with pytest.raises(ValueError):
        exp.shap_explain(n=50, X_ref=random_data)
    
    # X_ref > 1 row
    with pytest.raises(ValueError):
        exp.shap_explain(n=50, X_ref=long_data)
    
    exp.shap_explain(n=50)
    
    # Missing X names
    with pytest.raises(ValueError):
        exp.shap_single_point(X_new=random_data)
    
    # Missing X names
    with pytest.raises(ValueError):
        exp.shap_single_point(X_new=campaign.optimizer.X_best_f, X_ref=random_data)

    # Missing X names
    with pytest.raises(ValueError):
        exp.sensitivity(X_ref=random_data)
    
    # X_ref > 1 row
    with pytest.raises(ValueError):
        exp.sensitivity(X_ref=long_data)


if __name__ == '__main__':
    pytest.main([__file__, '-m', 'not slow'])



"""Custom implementations of acquisition functions using BoTorch API"""

from botorch.acquisition.monte_carlo import MCAcquisitionFunction
from botorch.sampling.normal import SobolQMCNormalSampler
from botorch.utils import t_batch_mode_transform
from botorch.acquisition.multi_objective.objective import IdentityMCMultiOutputObjective
from botorch.models.model import Model
from botorch.sampling.base import MCSampler
from botorch.acquisition.objective import MCAcquisitionObjective, PosteriorTransform

import torch
from torch import Tensor


class qMean(MCAcquisitionFunction):
    """
    Acquisition function which optimizes for the
    maximum value of the posterior mean
    """
    def __init__(self,
                 model: Model,
                 sampler: MCSampler | None = None,
                 objective: MCAcquisitionObjective | None = None,
                 posterior_transform: PosteriorTransform | None = None,
                 X_pending: Tensor | None = None):
        
        if sampler is None:
            sampler = SobolQMCNormalSampler(sample_shape=torch.Size([512]))
        
        # An objective is required if this aq is used on a multi-output model
        if objective is None:
            if model.num_outputs > 1:
                objective = IdentityMCMultiOutputObjective()
            else:
                objective = None  # This will be set to identity in super call

        super().__init__(model=model, sampler=sampler, objective=objective,
                         posterior_transform=posterior_transform, X_pending=X_pending)

    @t_batch_mode_transform()
    def forward(self,
                x: Tensor) -> Tensor:
        """
        Evaluate the acquisition function on the candidate set x
        """
        # x dimensions: b * q * d
        posterior = self.model.posterior(x)  # dimensions: b * q
        samples = self.get_posterior_samples(posterior)  # dimensions: n * b * q * o

        o = self.objective(samples, x)
        if self.objective._is_mo:
            o = o.sum(dim=-1)  # Sum over o if objective is MOO
        else:
            o = samples.sum(dim=-1)  # Sum over o if f no objectives

        # Mean over posterior, sum over q in any case
        return o.mean(dim=0).sum(dim=-1)


class qSpaceFill(MCAcquisitionFunction):
    """
    Acquisition function which optimizes for the
    maximum value of minimum distance between a point and the training data
    """
    def __init__(self,
                 model: Model,
                 X_baseline: Tensor,
                 sampler: MCSampler | None = None,
                 objective: MCAcquisitionObjective | None = None,
                 posterior_transform: PosteriorTransform | None = None,
                 X_pending: Tensor | None = None,):
        
        if sampler is None:
            sampler = SobolQMCNormalSampler(sample_shape=torch.Size([512]))
        
        # An objective is required if this aq is used on a multi-output model
        if model.num_outputs > 1:
            objective = IdentityMCMultiOutputObjective()
        else:
            objective = None

        super().__init__(model=model, sampler=sampler, objective=objective,
                         posterior_transform=posterior_transform, X_pending=X_pending)

        self.register_buffer('X_baseline', X_baseline)
        
    @t_batch_mode_transform()
    def forward(self,
                x: Tensor) -> Tensor:
        """
        Evaluate the acquisition function on the candidate set x
        """
        # x dimensions: b * q * d
        x_train = self.X_baseline
        
        # For sequential mode, add pending data points to "train"
        if self.X_pending is not None:
            x_train = torch.concat((x_train, self.X_pending))
               
        x = x.unsqueeze(-2)  # Insert expmt dimension before xdim
        x_train = x_train.unsqueeze(0).unsqueeze(0)  # Add batch and q dimensions
        
        # Norm distance over xdim
        dist = torch.norm(x-x_train, p=2, dim=-1)
        
        # Min over expmt and q (for joint evaluation)
        closest_dist = dist.min(dim=-1)[0].min(dim=-1)[0]

        return closest_dist



"""Obsidian branding colors and color maps"""

from matplotlib.colors import LinearSegmentedColormap


def hex_to_rgb(value: float):
    """Convert hex color to RGB tuple"""
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i + lv // 3], 16)/256 for i in range(0, lv, lv // 3))


class ColorScheme():
    """Color scheme for Obsidian"""
    pass


obsidian_colors = ColorScheme()


class Palette():
    """Individual color palettes"""
    pass


# Establish primary, secondary, and accent branding colors
primary = Palette()
primary.teal = '#00857C'
primary.white = '#FFFFFF'

secondary = Palette()
secondary.blue = '#0C2340'
secondary.light_teal = '#6ECEB2'
secondary.off_white = '#F7F7F7'

accent = Palette()
accent.lime = '#BFED33'
accent.lemon = '#FFF063'
accent.pastel_blue = '#69B8F7'
accent.vista_blue = '#688CE8'
accent.rich_blue = '#5450E4'

# Create repeating list of colors for easy access
obsidian_color_list = [primary.teal, secondary.blue, secondary.light_teal, accent.pastel_blue,
                       accent.lime, accent.vista_blue, accent.lemon, accent.rich_blue]

# Unused in palette, but used for color map
magenta = '#d04495'

obsidian_colors.primary = primary
obsidian_colors.secondary = secondary
obsidian_colors.accent = accent

# Wrap all colors also directly into obsidian_colors unsorted by type
# Allows the user to acecss colors directly from the obsidian_colors object
# Example: obsidian_colors.teal isntead of obsidian_colors.primary.teal
for attr in dir(obsidian_colors):
    if isinstance(getattr(obsidian_colors, attr), Palette):
        palette = getattr(obsidian_colors, attr)
        for sub_attr in dir(palette):
            if isinstance(getattr(palette, sub_attr), str):
                if '#' in getattr(palette, sub_attr):
                    setattr(obsidian_colors, sub_attr, getattr(palette, sub_attr))


class ColorMaps():
    """Continuous color maps from obsidian palettes"""
    pass


obsidian_cm = ColorMaps()

# Viridis = [blue, teal, lemon]
obsidian_cm.obsidian_viridis = LinearSegmentedColormap.from_list('obsidianViridis',
                                                                 colors=[hex_to_rgb(obsidian_colors.accent.rich_blue),
                                                                         hex_to_rgb(obsidian_colors.primary.teal),
                                                                         hex_to_rgb(obsidian_colors.accent.lemon)])

# Plasma = [blue, magenta, lemon]
obsidian_cm.obsidian_plasma = LinearSegmentedColormap.from_list('obsidianPlasma',
                                                                colors=[hex_to_rgb(obsidian_colors.accent.rich_blue),
                                                                        hex_to_rgb(magenta),
                                                                        hex_to_rgb(obsidian_colors.accent.lemon)])

# Mako = [blue, teal, light teal]
obsidian_cm.obsidian_mako = LinearSegmentedColormap.from_list('obsidianMako',
                                                              colors=[hex_to_rgb(obsidian_colors.blue),
                                                                      hex_to_rgb(obsidian_colors.teal),
                                                                      hex_to_rgb(obsidian_colors.light_teal)])

# Teal Shade = [light teal, teal]
obsidian_cm.obsidian_tealshade = LinearSegmentedColormap.from_list('obsidianTealShade',
                                                                   colors=[hex_to_rgb(obsidian_colors.light_teal),
                                                                           hex_to_rgb(obsidian_colors.teal)])

# Blue Shade = [vista blue, rich blue]
obsidian_cm.obsidian_blueshade = LinearSegmentedColormap.from_list('obsidianBlueShade',
                                                                   colors=[hex_to_rgb(obsidian_colors.vista_blue),
                                                                           hex_to_rgb(obsidian_colors.rich_blue)])

obsidian_colors.cm = obsidian_cm
